@@ -13,12 +13,17 @@
 // limitations under the License.
 
 #![deny(clippy::all)]
+use std::cell::RefCell;
+use std::cell::RefMut;
 use std::collections::BTreeMap;
 use std::collections::BTreeSet;
 use std::ffi::OsStr;
+use std::fs::File;
 use std::io::BufRead;
 use std::io::BufReader;
 use std::io::ErrorKind;
+use std::os::fd::AsRawFd;
+use std::os::fd::BorrowedFd;
 use std::path::Path;
 use std::path::PathBuf;
 use std::str::FromStr;
@@ -36,6 +41,8 @@ pub use types::*;
 #[cfg(test)]
 mod test;
 
+use common::util;
+
 pub const DEFAULT_CG_ROOT: &str = "/sys/fs/cgroup";
 
 #[derive(Error, Debug)]
@@ -57,6 +64,7 @@ pub type Result<T> = std::result::Result<T, Error>;
 pub struct CgroupReader {
     relative_path: PathBuf,
     dir: Dir,
+    buffer: RefCell<Vec<u8>>,
 }
 
 fn parse_integer_or_max(s: &str) -> std::result::Result<i64, String> {
@@ -235,7 +243,7 @@ impl CgroupReader {
         relative_path: PathBuf,
         validate: bool,
     ) -> Result<CgroupReader> {
-        let mut path = root.clone();
+        let mut path = root;
         match relative_path.strip_prefix("/") {
             Ok(p) => path.push(p),
             _ => path.push(&relative_path),
@@ -244,7 +252,10 @@ impl CgroupReader {
 
         // Check that it's a cgroup2 fs
         if validate {
-            let statfs = match fstatfs(&dir) {
+            // SAFETY: Fix when https://github.com/nix-rust/nix/issues/2546 is
+            let dir = unsafe { BorrowedFd::borrow_raw(dir.as_raw_fd()) };
+
+            let statfs = match fstatfs(dir) {
                 Ok(s) => s,
                 Err(e) => {
                     return Err(Error::IoError(
@@ -259,7 +270,11 @@ impl CgroupReader {
             }
         }
 
-        Ok(CgroupReader { relative_path, dir })
+        Ok(CgroupReader {
+            relative_path,
+            dir,
+            buffer: RefCell::new(Vec::new()),
+        })
     }
 
     pub fn root() -> Result<CgroupReader> {
@@ -272,12 +287,21 @@ impl CgroupReader {
         &self.relative_path
     }
 
+    fn read_file_to_str(
+        &self,
+        file_name: impl AsRef<Path>,
+        file: &File,
+    ) -> Result<RefMut<'_, str>> {
+        util::read_kern_file_to_internal_buffer(&self.buffer, file)
+            .map_err(|e| self.io_error(file_name.as_ref(), e))
+    }
+
     pub fn read_inode_number(&self) -> Result<u64> {
         let meta = self
             .dir
             .metadata(".")
             .map_err(|e| self.io_error(self.dir.recover_path().unwrap_or_else(|_| "".into()), e))?;
-        Ok(meta.stat().st_ino as u64)
+        Ok(meta.stat().st_ino)
     }
 
     /// Read a value from a file that has a single line. If the file is empty,
@@ -287,14 +311,10 @@ impl CgroupReader {
             .dir
             .open_file(file_name)
             .map_err(|e| self.io_error(file_name, e))?;
-        let buf_reader = BufReader::new(file);
-        let line = buf_reader
-            .lines()
-            .next()
-            .unwrap_or_else(|| Ok("".to_owned()));
-        let line = line.map_err(|e| self.io_error(file_name, e))?;
+        let content = self.read_file_to_str(file_name, &file)?;
+        let line = content.lines().next().unwrap_or("");
         line.parse::<T>()
-            .map_err(move |_| self.unexpected_line(file_name, line))
+            .map_err(move |_| self.unexpected_line(file_name, line.to_string()))
     }
 
     /// Read a value from a file that has a single line. If the file is empty,
@@ -304,12 +324,11 @@ impl CgroupReader {
             .dir
             .open_file(file_name)
             .map_err(|e| self.io_error(file_name, e))?;
-        let buf_reader = BufReader::new(file);
-        if let Some(line) = buf_reader.lines().next() {
-            let line = line.map_err(|e| self.io_error(file_name, e))?;
+        let content = self.read_file_to_str(file_name, &file)?;
+        if let Some(line) = content.lines().next() {
             return line
                 .parse::<T>()
-                .map_err(move |_| self.unexpected_line(file_name, line));
+                .map_err(move |_| self.unexpected_line(file_name, line.to_string()));
         }
         Err(self.invalid_file_format(file_name))
     }
@@ -344,6 +363,22 @@ impl CgroupReader {
         self.read_singleline_controllers("cgroup.subtree_control")
     }
 
+    /// Read pids.current - returning current cgroup number of processes
+    pub fn read_pids_current(&self) -> Result<u64> {
+        self.read_singleline_file("pids.current")
+    }
+
+    /// Read pids.max - returning max cgroup number of processes
+    pub fn read_pids_max(&self) -> Result<i64> {
+        self.read_singleline_integer_or_max_stat_file("pids.max")
+    }
+
+    /// Read memory.min - returning memory.min limit in bytes
+    /// Will return -1 if the content is max
+    pub fn read_memory_min(&self) -> Result<i64> {
+        self.read_singleline_integer_or_max_stat_file("memory.min")
+    }
+
     /// Read memory.low - returning memory.low limit in bytes
     /// Will return -1 if the content is max
     pub fn read_memory_low(&self) -> Result<i64> {
@@ -394,22 +429,26 @@ impl CgroupReader {
 
     /// Read cpu.stat - returning assorted cpu consumption statistics
     pub fn read_cpu_stat(&self) -> Result<CpuStat> {
-        CpuStat::read(&self)
+        CpuStat::read(self)
     }
 
     /// Read io.stat - returning assorted io consumption statistics
     pub fn read_io_stat(&self) -> Result<BTreeMap<String, IoStat>> {
-        IoStat::read(&self, "io.stat")
+        IoStat::read(self, "io.stat")
     }
 
     /// Read memory.stat - returning assorted memory consumption
     /// statistics
     pub fn read_memory_stat(&self) -> Result<MemoryStat> {
-        MemoryStat::read(&self)
+        MemoryStat::read(self)
     }
 
     pub fn read_memory_events(&self) -> Result<MemoryEvents> {
-        MemoryEvents::read(&self)
+        MemoryEvents::read(self)
+    }
+
+    pub fn read_memory_events_local(&self) -> Result<MemoryEventsLocal> {
+        MemoryEventsLocal::read(self)
     }
 
     pub fn read_cgroup_stat(&self) -> Result<CgroupStat> {
@@ -485,30 +524,29 @@ impl CgroupReader {
             .dir
             .open_file(file_name)
             .map_err(|e| self.io_error(file_name, e))?;
-        let buf_reader = BufReader::new(file);
-        for line in buf_reader.lines() {
-            let line = line.map_err(|e| self.io_error(file_name, e))?;
-            let items = line.split_whitespace().collect::<Vec<_>>();
+        let content = self.read_file_to_str(file_name, &file)?;
+        for line in content.lines() {
+            let items = line.split_ascii_whitespace().collect::<Vec<_>>();
             // Need to have at least the field name + at least one N0=val item
             if items.len() < 2 {
-                return Err(self.unexpected_line(file_name, line));
+                return Err(self.unexpected_line(file_name, line.to_string()));
             }
             let field = items[0];
             for item in items.iter().skip(1) {
                 let kv = item.split('=').collect::<Vec<_>>();
                 if kv.len() != 2 || kv[0].len() < 2 || !kv[0].starts_with('N') {
-                    return Err(self.unexpected_line(file_name, line));
+                    return Err(self.unexpected_line(file_name, line.to_string()));
                 }
                 let mut kchars = kv[0].chars();
                 kchars.next();
                 let numa_node = kchars
                     .as_str()
                     .parse::<u32>()
-                    .map_err(|_| self.unexpected_line(file_name, line.clone()))?;
+                    .map_err(|_| self.unexpected_line(file_name, line.to_string()))?;
                 parse_and_set_fields!(
                     s.entry(numa_node).or_default();
                     field;
-                    kv[1].parse().map_err(|_| self.unexpected_line(file_name, line.clone()))?;
+                    kv[1].parse().map_err(|_| self.unexpected_line(file_name, line.to_string()))?;
                     [
                         anon,
                         file,
@@ -568,7 +606,11 @@ impl CgroupReader {
                     };
                     let mut relative_path = self.relative_path.clone();
                     relative_path.push(entry.file_name());
-                    Some(CgroupReader { relative_path, dir })
+                    Some(CgroupReader {
+                        relative_path,
+                        dir,
+                        buffer: RefCell::new(Vec::new()),
+                    })
                 }
                 _ => None,
             }))
@@ -618,13 +660,14 @@ macro_rules! key_values_format {
                 let buf_reader = BufReader::new(file);
                 for line in buf_reader.lines() {
                     let line = line.map_err(|e| r.io_error(file_name, e))?;
-                    let items = line.split_whitespace().collect::<Vec<_>>();
-                    if items.len() != 2 {
+                    let mut items = line.split_ascii_whitespace();
+                    let key = items.next().ok_or_else(|| r.unexpected_line(file_name, line.clone()))?;
+                    let val_str = items.next().ok_or_else(|| r.unexpected_line(file_name, line.clone()))?;
+                    if items.next().is_some() {
                         return Err(r.unexpected_line(file_name, line));
                     }
-                    let key = items[0];
-                    let val = items[1].parse::<_>().map_err(|_| r.unexpected_line(file_name, line.clone()))?;
-                    match key.as_ref() {
+                    let val = val_str.parse::<_>().map_err(|_| r.unexpected_line(file_name, line.clone()))?;
+                    match key {
                         $(stringify!($field) => s.$field = Some(val),)*
                         _ => (),
                     };
@@ -651,10 +694,13 @@ key_values_format!(CpuStat; cpu.stat; [
 key_values_format!(MemoryStat; memory.stat; [
     anon,
     file,
+    kernel,
     kernel_stack,
     slab,
     sock,
     shmem,
+    zswap,
+    zswapped,
     file_mapped,
     file_dirty,
     file_writeback,
@@ -694,6 +740,14 @@ key_values_format!(MemoryEvents; memory.events; [
     oom_kill
 ]);
 
+key_values_format!(MemoryEventsLocal; memory.events.local; [
+    low,
+    high,
+    max,
+    oom,
+    oom_kill
+]);
+
 key_values_format!(CgroupStat; cgroup.stat; [nr_descendants, nr_dying_descendants]);
 
 // Trait to add a read() method for `<string> key=value` formatted files
@@ -727,31 +781,36 @@ macro_rules! name_key_equal_value_format {
                         }
                         r.io_error(file_name.clone(), e)
                     })?;
-                    let items = line.split_whitespace().collect::<Vec<_>>();
                     // as an example, io.stat looks like:
+                    // 7:0
                     // 253:0 rbytes=531745786880 wbytes=1623798909952 ...
+                    let mut items = line.split_terminator(' ').peekable();
+                    let Some(first_item) = items.next() else {
+                      continue
+                    };
+                    // Skip lines with no kv pairs
+                    if items.peek().is_none() {
+                      continue
+                    }
                     let mut s = $struct::default();
-                    for item in items.iter().skip(1) {
-                        let kv = item.split("=").collect::<Vec<_>>();
-                        if kv.len() != 2 {
-                            return Err(r.invalid_file_format(file_name));
-                        }
-                        // Certain keys such as cost.usage can not be struct fields so must use cost_usage
-                        let key = kv[0].replace(".", "_");
+                    for item in items {
+                        let (key, value) = item.split_once('=').ok_or_else(|| r.invalid_file_format(file_name.clone()))?;
+                        // Certain keys such as cost.usage cannot be struct fields so must use cost_usage
+                        let key = key.replace(".", "_");
                         parse_and_set_fields!(
                             s;
                             key.as_ref();
-                            kv[1].parse().map_err(|_| r.unexpected_line(file_name.clone(), line.clone()))?;
+                            value.parse().map_err(|_| r.unexpected_line(file_name.clone(), line.clone()))?;
                             [ $($field,)* ]
                         )
-                    };
-                    if s == $struct::default() {
-                        return Err(r.invalid_file_format(file_name))
                     }
-                    map.insert(items[0].to_string(), s);
+                    if s == $struct::default() {
+                        return Err(r.invalid_file_format(file_name));
+                    }
+                    map.insert(first_item.to_string(), s);
                 }
                 if !$allows_empty.0 && map.is_empty() {
-                     Err(r.invalid_file_format(file_name))
+                    Err(r.invalid_file_format(file_name))
                 } else {
                     Ok(map)
                 }
