@@ -1,5 +1,4 @@
-#![allow(clippy::let_unit_value)]
-#![warn(clippy::absolute_paths)]
+//! End-to-end tests for `libbpf-rs`.
 
 mod common;
 
@@ -29,6 +28,8 @@ use std::time::Duration;
 use libbpf_rs::num_possible_cpus;
 use libbpf_rs::AsRawLibbpf;
 use libbpf_rs::Iter;
+use libbpf_rs::KprobeMultiOpts;
+use libbpf_rs::KprobeOpts;
 use libbpf_rs::Linker;
 use libbpf_rs::MapCore;
 use libbpf_rs::MapFlags;
@@ -40,6 +41,8 @@ use libbpf_rs::ObjectBuilder;
 use libbpf_rs::Program;
 use libbpf_rs::ProgramInput;
 use libbpf_rs::ProgramType;
+use libbpf_rs::RawTracepointOpts;
+use libbpf_rs::TracepointCategory;
 use libbpf_rs::TracepointOpts;
 use libbpf_rs::UprobeOpts;
 use libbpf_rs::UsdtOpts;
@@ -50,10 +53,10 @@ use scopeguard::defer;
 use tempfile::NamedTempFile;
 use test_tag::tag;
 
-use crate::common::bump_rlimit_mlock;
 use crate::common::get_map;
 use crate::common::get_map_mut;
 use crate::common::get_prog_mut;
+use crate::common::get_symbol_offset;
 use crate::common::get_test_object;
 use crate::common::get_test_object_path;
 use crate::common::open_test_object;
@@ -62,8 +65,6 @@ use crate::common::with_ringbuffer;
 #[tag(root)]
 #[test]
 fn test_object_build_and_load() {
-    bump_rlimit_mlock();
-
     get_test_object("runqslower.bpf.o");
 }
 
@@ -127,8 +128,6 @@ fn test_object_name() {
 #[tag(root)]
 #[test]
 fn test_object_maps() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let _map = get_map_mut(&mut obj, "start");
     let _map = get_map_mut(&mut obj, "events");
@@ -138,8 +137,6 @@ fn test_object_maps() {
 #[tag(root)]
 #[test]
 fn test_object_maps_iter() {
-    bump_rlimit_mlock();
-
     let obj = get_test_object("runqslower.bpf.o");
     for map in obj.maps() {
         eprintln!("{:?}", map.name());
@@ -151,8 +148,6 @@ fn test_object_maps_iter() {
 #[tag(root)]
 #[test]
 fn test_object_map_key_value_size() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
 
@@ -167,8 +162,6 @@ fn test_object_map_key_value_size() {
 #[tag(root)]
 #[test]
 fn test_object_map_update_batch() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
 
@@ -256,8 +249,6 @@ fn test_object_map_update_batch() {
 #[tag(root)]
 #[test]
 fn test_object_map_lookup_batch() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
     let data = HashMap::from([
@@ -318,8 +309,6 @@ fn test_object_map_lookup_batch() {
 #[tag(root)]
 #[test]
 fn test_object_map_delete_batch() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
 
@@ -365,7 +354,6 @@ fn test_object_map_delete_batch() {
 #[tag(root)]
 #[test]
 pub fn test_map_info() {
-    #[allow(clippy::needless_update)]
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
         map_flags: libbpf_sys::BPF_ANY,
@@ -403,8 +391,6 @@ pub fn test_map_info() {
 #[tag(root)]
 #[test]
 fn test_object_percpu_lookup() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("percpu_map.bpf.o");
     let map = get_map_mut(&mut obj, "percpu_map");
     let res = map
@@ -422,8 +408,6 @@ fn test_object_percpu_lookup() {
 #[tag(root)]
 #[test]
 fn test_object_percpu_invalid_lookup_fn() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("percpu_map.bpf.o");
     let map = get_map_mut(&mut obj, "percpu_map");
 
@@ -433,8 +417,6 @@ fn test_object_percpu_invalid_lookup_fn() {
 #[tag(root)]
 #[test]
 fn test_object_percpu_update() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("percpu_map.bpf.o");
     let map = get_map_mut(&mut obj, "percpu_map");
     let key = (0_u32).to_ne_bytes();
@@ -458,8 +440,6 @@ fn test_object_percpu_update() {
 #[tag(root)]
 #[test]
 fn test_object_percpu_invalid_update_fn() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("percpu_map.bpf.o");
     let map = get_map_mut(&mut obj, "percpu_map");
     let key = (0_u32).to_ne_bytes();
@@ -472,8 +452,6 @@ fn test_object_percpu_invalid_update_fn() {
 #[tag(root)]
 #[test]
 fn test_object_percpu_lookup_update() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("percpu_map.bpf.o");
     let map = get_map_mut(&mut obj, "percpu_map");
     let key = (0_u32).to_ne_bytes();
@@ -501,8 +479,6 @@ fn test_object_percpu_lookup_update() {
 #[tag(root)]
 #[test]
 fn test_object_map_empty_lookup() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
 
@@ -516,8 +492,6 @@ fn test_object_map_empty_lookup() {
 #[tag(root)]
 #[test]
 fn test_object_map_queue_crud() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("tracepoint.bpf.o");
     let queue = get_map_mut(&mut obj, "queue");
 
@@ -564,8 +538,6 @@ fn test_object_map_queue_crud() {
 #[tag(root)]
 #[test]
 fn test_object_map_bloom_filter_crud() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("tracepoint.bpf.o");
     let bloom_filter = get_map_mut(&mut obj, "bloom_filter");
 
@@ -615,8 +587,6 @@ fn test_object_map_bloom_filter_crud() {
 #[tag(root)]
 #[test]
 fn test_object_map_stack_crud() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("tracepoint.bpf.o");
     let stack = get_map_mut(&mut obj, "stack");
 
@@ -662,8 +632,6 @@ fn test_object_map_stack_crud() {
 #[tag(root)]
 #[test]
 fn test_object_map_mutation() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
     start
@@ -687,8 +655,6 @@ fn test_object_map_mutation() {
 #[tag(root)]
 #[test]
 fn test_object_map_lookup_flags() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
     start
@@ -702,8 +668,6 @@ fn test_object_map_lookup_flags() {
 #[tag(root)]
 #[test]
 fn test_object_map_key_iter() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
 
@@ -734,8 +698,6 @@ fn test_object_map_key_iter() {
 #[tag(root)]
 #[test]
 fn test_object_map_key_iter_empty() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let start = get_map_mut(&mut obj, "start");
     let mut count = 0;
@@ -748,8 +710,6 @@ fn test_object_map_key_iter_empty() {
 #[tag(root)]
 #[test]
 fn test_object_map_pin() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let mut map = get_map_mut(&mut obj, "start");
     let path = "/sys/fs/bpf/mymap_test_object_map_pin";
@@ -768,8 +728,6 @@ fn test_object_map_pin() {
 #[tag(root)]
 #[test]
 fn test_object_loading_pinned_map_from_path() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let mut map = get_map_mut(&mut obj, "start");
     let path = "/sys/fs/bpf/mymap_test_pin_to_load_from_path";
@@ -789,8 +747,6 @@ fn test_object_loading_pinned_map_from_path() {
 #[tag(root)]
 #[test]
 fn test_program_loading_fd_from_pinned_path() {
-    bump_rlimit_mlock();
-
     let path = "/sys/fs/bpf/myprog_test_pin_to_load_from_path";
     let prog_name = "handle__sched_switch";
 
@@ -812,8 +768,6 @@ fn test_program_loading_fd_from_pinned_path() {
 #[tag(root)]
 #[test]
 fn test_program_loading_fd_from_pinned_path_with_wrong_pin_type() {
-    bump_rlimit_mlock();
-
     let path = "/sys/fs/bpf/mymap_test_pin_to_load_from_path";
     let map_name = "events";
 
@@ -822,7 +776,7 @@ fn test_program_loading_fd_from_pinned_path_with_wrong_pin_type() {
     map.pin(path).expect("pinning map failed");
 
     // Must fail, as the pinned path points to a map, not program.
-    let _ = Program::fd_from_pinned_path(path).expect_err("program fd obtained from pinned map");
+    let _err = Program::fd_from_pinned_path(path).expect_err("program fd obtained from pinned map");
 
     map.unpin(path).expect("unpinning program failed");
 }
@@ -830,8 +784,6 @@ fn test_program_loading_fd_from_pinned_path_with_wrong_pin_type() {
 #[tag(root)]
 #[test]
 fn test_object_loading_loaded_map_from_id() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let map = get_map_mut(&mut obj, "start");
     let id = map.info().expect("to get info from map 'start'").info.id;
@@ -848,8 +800,6 @@ fn test_object_loading_loaded_map_from_id() {
 #[tag(root)]
 #[test]
 fn test_object_programs() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let _prog = get_prog_mut(&mut obj, "handle__sched_wakeup");
     let _prog = get_prog_mut(&mut obj, "handle__sched_wakeup_new");
@@ -860,8 +810,6 @@ fn test_object_programs() {
 #[tag(root)]
 #[test]
 fn test_object_programs_iter_mut() {
-    bump_rlimit_mlock();
-
     let obj = get_test_object("runqslower.bpf.o");
     assert!(obj.progs().count() == 3);
 }
@@ -869,8 +817,6 @@ fn test_object_programs_iter_mut() {
 #[tag(root)]
 #[test]
 fn test_object_program_pin() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let mut prog = get_prog_mut(&mut obj, "handle__sched_wakeup");
     let path = "/sys/fs/bpf/myprog";
@@ -885,7 +831,7 @@ fn test_object_program_pin() {
 
     // Backup cleanup method in case test errors
     defer! {
-        let _ = fs::remove_file(path);
+        let _unused = fs::remove_file(path);
     }
 
     // Unpin should be successful
@@ -896,8 +842,6 @@ fn test_object_program_pin() {
 #[tag(root)]
 #[test]
 fn test_object_link_pin() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sched_wakeup");
     let mut link = prog.attach().expect("failed to attach prog");
@@ -914,7 +858,7 @@ fn test_object_link_pin() {
 
     // Backup cleanup method in case test errors
     defer! {
-        let _ = fs::remove_file(path);
+        let _unused = fs::remove_file(path);
     }
 
     // Unpin should be successful
@@ -925,8 +869,6 @@ fn test_object_link_pin() {
 #[tag(root)]
 #[test]
 fn test_object_reuse_pined_map() {
-    bump_rlimit_mlock();
-
     let path = "/sys/fs/bpf/mymap_test_object_reuse_pined_map";
     let key = vec![1, 2, 3, 4];
     let val = vec![1, 2, 3, 4, 5, 6, 7, 8];
@@ -945,7 +887,7 @@ fn test_object_reuse_pined_map() {
 
     // Backup cleanup method in case test errors somewhere
     defer! {
-        let _ = fs::remove_file(path);
+        let _unused = fs::remove_file(path);
     }
 
     // Reuse the pinned map
@@ -976,8 +918,6 @@ fn test_object_reuse_pined_map() {
 #[tag(root)]
 #[test]
 fn test_object_ringbuf_raw() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1046,13 +986,28 @@ fn test_object_ringbuf_raw() {
     // Consume from a (potentially) empty ring buffer using poll()
     let ret = mgr.poll_raw(Duration::from_millis(100));
     assert!(ret >= 0);
+
+    // Call getpid multiple times, to refill the ring buffer.
+    for _ in 1..=10 {
+        unsafe { libc::getpid() };
+    }
+
+    // Consume exactly one item
+    let ret = mgr.consume_raw_n(1);
+    assert!(ret == 1);
+
+    // Consume two items
+    let ret = mgr.consume_raw_n(2);
+    assert!(ret == 2);
+
+    // Consume all the remaining items, but no more than 10
+    let ret = mgr.consume_raw_n(10);
+    assert!((7..=10).contains(&ret));
 }
 
 #[tag(root)]
 #[test]
 fn test_object_ringbuf_err_callback() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1113,8 +1068,6 @@ fn test_object_ringbuf_err_callback() {
 #[tag(root)]
 #[test]
 fn test_object_ringbuf() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1192,8 +1145,6 @@ fn test_object_ringbuf() {
 #[tag(root)]
 #[test]
 fn test_object_ringbuf_closure() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1262,8 +1213,6 @@ fn test_object_ringbuf_closure() {
 #[tag(root)]
 #[test]
 fn test_object_ringbuf_with_closed_map() {
-    bump_rlimit_mlock();
-
     fn test(poll_fn: impl FnOnce(&libbpf_rs::RingBuffer)) {
         let mut value = 0i32;
 
@@ -1271,7 +1220,7 @@ fn test_object_ringbuf_with_closed_map() {
             let mut obj = get_test_object("tracepoint.bpf.o");
             let prog = get_prog_mut(&mut obj, "handle__tracepoint");
             let _link = prog
-                .attach_tracepoint("syscalls", "sys_enter_getpid")
+                .attach_tracepoint(TracepointCategory::Syscalls, "sys_enter_getpid")
                 .expect("failed to attach prog");
 
             let map = get_map_mut(&mut obj, "ringbuf");
@@ -1316,8 +1265,6 @@ fn test_object_user_ringbuf() {
 
     unsafe impl Plain for MyStruct {}
 
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("user_ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1357,8 +1304,6 @@ fn test_object_user_ringbuf() {
 #[tag(root)]
 #[test]
 fn test_object_user_ringbuf_reservation_too_big() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("user_ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
@@ -1374,14 +1319,12 @@ fn test_object_user_ringbuf_reservation_too_big() {
 #[tag(root)]
 #[test]
 fn test_object_user_ringbuf_not_enough_space() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("user_ringbuf.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sys_enter_getpid");
     let _link = prog.attach().expect("failed to attach prog");
     let urb_map = get_map_mut(&mut obj, "user_ringbuf");
     let user_ringbuf = UserRingBuffer::new(&urb_map).expect("failed to create user ringbuf");
-    let _ = user_ringbuf
+    let _sample = user_ringbuf
         .reserve(1024 * 3)
         .expect("failed to reserve space");
     let err = user_ringbuf.reserve(1024 * 3).unwrap_err();
@@ -1395,8 +1338,6 @@ fn test_object_user_ringbuf_not_enough_space() {
 #[tag(root)]
 #[test]
 fn test_object_task_iter() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("taskiter.bpf.o");
     let prog = get_prog_mut(&mut obj, "dump_pid");
     let link = prog.attach().expect("failed to attach prog");
@@ -1431,8 +1372,6 @@ fn test_object_task_iter() {
 #[tag(root)]
 #[test]
 fn test_object_map_iter() {
-    bump_rlimit_mlock();
-
     // Create a map for iteration test.
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
@@ -1483,8 +1422,6 @@ fn test_object_map_iter() {
 #[tag(root)]
 #[test]
 fn test_object_map_create_and_pin() {
-    bump_rlimit_mlock();
-
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
         map_flags: libbpf_sys::BPF_F_NO_PREALLOC,
@@ -1529,9 +1466,6 @@ fn test_object_map_create_and_pin() {
 #[tag(root)]
 #[test]
 fn test_object_map_create_without_name() {
-    bump_rlimit_mlock();
-
-    #[allow(clippy::needless_update)]
     let opts = libbpf_sys::bpf_map_create_opts {
         sz: size_of::<libbpf_sys::bpf_map_create_opts>() as libbpf_sys::size_t,
         map_flags: libbpf_sys::BPF_F_NO_PREALLOC,
@@ -1563,12 +1497,10 @@ fn test_object_map_create_without_name() {
     assert_eq!(val, res);
 }
 
-/// Test whether we can obtain multiple `MapHandle`s from a `Map
+/// Test whether we can obtain multiple `MapHandle`s from a `Map`.
 #[tag(root)]
 #[test]
 fn test_object_map_handle_clone() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let map = get_map_mut(&mut obj, "events");
     let handle1 = MapHandle::try_from(&map).expect("failed to create handle from Map");
@@ -1576,12 +1508,14 @@ fn test_object_map_handle_clone() {
     assert_eq!(map.map_type(), handle1.map_type());
     assert_eq!(map.key_size(), handle1.key_size());
     assert_eq!(map.value_size(), handle1.value_size());
+    assert_eq!(map.max_entries(), handle1.max_entries());
 
     let handle2 = MapHandle::try_from(&handle1).expect("failed to duplicate existing handle");
     assert_eq!(handle1.name(), handle2.name());
     assert_eq!(handle1.map_type(), handle2.map_type());
     assert_eq!(handle1.key_size(), handle2.key_size());
     assert_eq!(handle1.value_size(), handle2.value_size());
+    assert_eq!(handle1.max_entries(), handle2.max_entries());
 
     let info1 = map.info().expect("failed to get map info from map");
     let info2 = handle2.info().expect("failed to get map info from handle");
@@ -1594,8 +1528,6 @@ fn test_object_map_handle_clone() {
 #[tag(root)]
 #[test]
 fn test_object_usdt() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("usdt.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__usdt");
 
@@ -1622,8 +1554,6 @@ fn test_object_usdt() {
 #[tag(root)]
 #[test]
 fn test_object_usdt_cookie() {
-    bump_rlimit_mlock();
-
     let cookie_val = 1337u16;
     let mut obj = get_test_object("usdt.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__usdt_with_cookie");
@@ -1655,8 +1585,6 @@ fn test_object_usdt_cookie() {
 #[tag(root)]
 #[test]
 fn test_map_probes() {
-    bump_rlimit_mlock();
-
     let supported = MapType::Array
         .is_supported()
         .expect("failed to query if Array map is supported");
@@ -1668,8 +1596,6 @@ fn test_map_probes() {
 #[tag(root)]
 #[test]
 fn test_program_probes() {
-    bump_rlimit_mlock();
-
     let supported = ProgramType::SocketFilter
         .is_supported()
         .expect("failed to query if SocketFilter program is supported");
@@ -1681,8 +1607,6 @@ fn test_program_probes() {
 #[tag(root)]
 #[test]
 fn test_program_helper_probes() {
-    bump_rlimit_mlock();
-
     let supported = ProgramType::SocketFilter
         .is_helper_supported(libbpf_sys::BPF_FUNC_map_lookup_elem)
         .expect("failed to query if helper supported");
@@ -1699,8 +1623,6 @@ fn test_program_helper_probes() {
 #[tag(root)]
 #[test]
 fn test_object_open_program_insns() {
-    bump_rlimit_mlock();
-
     let open_obj = open_test_object("usdt.bpf.o");
     let prog = open_obj
         .progs()
@@ -1714,24 +1636,92 @@ fn test_object_open_program_insns() {
 #[tag(root)]
 #[test]
 fn test_object_program_insns() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("usdt.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__usdt");
     let insns = prog.insns();
     assert!(!insns.is_empty());
 }
 
+/// Check that we can attach a BPF program to a kernel kprobe.
+#[tag(root)]
+#[test]
+fn test_object_kprobe() {
+    let mut obj = get_test_object("kprobe.bpf.o");
+    let prog = get_prog_mut(&mut obj, "handle__kprobe");
+    let _link = prog
+        .attach_kprobe(false, "bpf_fentry_test1")
+        .expect("failed to attach prog");
+}
+
+/// Check that we can attach a BPF program to a kernel kprobe, providing
+/// additional options.
+#[tag(root)]
+#[test]
+fn test_object_kprobe_with_opts() {
+    let mut obj = get_test_object("kprobe.bpf.o");
+    let prog = get_prog_mut(&mut obj, "handle__kprobe");
+    let opts = KprobeOpts::default();
+    let _link = prog
+        .attach_kprobe_with_opts(false, "bpf_fentry_test1", opts)
+        .expect("failed to attach prog");
+}
+
+/// Check that we can attach a BPF program to multiple kernel kprobes using
+/// `kprobe_multi`.
+#[tag(root)]
+#[test]
+#[ignore = "requires kernel with kprobe multi support"]
+fn test_object_kprobe_multi() {
+    let mut open_obj = open_test_object("kprobe.bpf.o");
+    open_obj
+        .progs_mut()
+        .find(|prog| prog.name() == "handle__kprobe")
+        .expect("failed to find `handle__kprobe` program")
+        .set_attach_type(libbpf_rs::ProgramAttachType::KprobeMulti);
+
+    let mut obj = open_obj.load().expect("failed to load object");
+    let prog = get_prog_mut(&mut obj, "handle__kprobe");
+    let _link = prog
+        .attach_kprobe_multi(false, vec!["bpf_fentry_test1", "bpf_fentry_test2"])
+        .expect("failed to attach prog");
+}
+
+/// Check that we can attach a BPF program to multiple kernel kprobes using
+/// `kprobe_multi`, providing additional options.
+#[tag(root)]
+#[test]
+#[ignore = "requires kernel with kprobe multi support"]
+fn test_object_kprobe_multi_with_opts() {
+    let mut open_obj = open_test_object("kprobe.bpf.o");
+    open_obj
+        .progs_mut()
+        .find(|prog| prog.name() == "handle__kprobe")
+        .expect("failed to find `handle__kprobe` program")
+        .set_attach_type(libbpf_rs::ProgramAttachType::KprobeMulti);
+
+    let mut obj = open_obj.load().expect("failed to load object");
+    let prog = get_prog_mut(&mut obj, "handle__kprobe");
+
+    let opts = KprobeMultiOpts {
+        symbols: vec![
+            "bpf_fentry_test1".to_string(),
+            "bpf_fentry_test2".to_string(),
+        ],
+        ..Default::default()
+    };
+    let _link = prog
+        .attach_kprobe_multi_with_opts(opts)
+        .expect("failed to attach prog");
+}
+
 /// Check that we can attach a BPF program to a kernel tracepoint.
 #[tag(root)]
 #[test]
 fn test_object_tracepoint() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("tracepoint.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__tracepoint");
     let _link = prog
-        .attach_tracepoint("syscalls", "sys_enter_getpid")
+        .attach_tracepoint(TracepointCategory::Syscalls, "sys_enter_getpid")
         .expect("failed to attach prog");
 
     let map = get_map_mut(&mut obj, "ringbuf");
@@ -1748,8 +1738,6 @@ fn test_object_tracepoint() {
 #[tag(root)]
 #[test]
 fn test_object_tracepoint_with_opts() {
-    bump_rlimit_mlock();
-
     let cookie_val = 42u16;
     let mut obj = get_test_object("tracepoint.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__tracepoint_with_cookie");
@@ -1759,7 +1747,68 @@ fn test_object_tracepoint_with_opts() {
         ..TracepointOpts::default()
     };
     let _link = prog
-        .attach_tracepoint_with_opts("syscalls", "sys_enter_getpid", opts)
+        .attach_tracepoint_with_opts(TracepointCategory::Syscalls, "sys_enter_getpid", opts)
+        .expect("failed to attach prog");
+
+    let map = get_map_mut(&mut obj, "ringbuf");
+    let action = || {
+        let _pid = unsafe { libc::getpid() };
+    };
+    let result = with_ringbuffer(&map, action);
+
+    assert_eq!(result, cookie_val.into());
+}
+
+/// Check that we can attach a BPF program to a kernel raw tracepoint.
+#[tag(root)]
+#[test]
+fn test_object_raw_tracepoint() {
+    let mut open_obj = open_test_object("tracepoint.bpf.o");
+    open_obj
+        .progs_mut()
+        .find(|prog| prog.name() == "handle__tracepoint")
+        .expect("failed to find `handle__tracepoint` program")
+        .set_prog_type(libbpf_rs::ProgramType::RawTracepoint);
+
+    let mut obj = open_obj.load().expect("failed to load object");
+    let prog = get_prog_mut(&mut obj, "handle__tracepoint");
+    let _link = prog
+        .attach_raw_tracepoint("sys_enter")
+        .expect("failed to attach prog");
+
+    let map = get_map_mut(&mut obj, "ringbuf");
+    let action = || {
+        let _pid = unsafe { libc::getpid() };
+    };
+    let result = with_ringbuffer(&map, action);
+
+    assert_eq!(result, 1);
+}
+
+/// Check that we can attach a BPF program to a kernel raw tracepoint, providing
+/// additional options.
+#[tag(root)]
+#[test]
+#[ignore = "requires kernel with bpf_get_attach_cookie for raw tracepoints"]
+fn test_object_raw_tracepoint_with_opts() {
+    let cookie_val = 42u16;
+
+    let mut open_obj = open_test_object("tracepoint.bpf.o");
+    open_obj
+        .progs_mut()
+        .find(|prog| prog.name() == "handle__tracepoint_with_cookie")
+        .expect("failed to find `handle__tracepoint` program")
+        .set_prog_type(libbpf_rs::ProgramType::RawTracepoint);
+
+    let mut obj = open_obj.load().expect("failed to load object");
+    let prog = get_prog_mut(&mut obj, "handle__tracepoint_with_cookie");
+
+    let opts = RawTracepointOpts {
+        cookie: cookie_val.into(),
+        ..Default::default()
+    };
+    let _link = prog
+        .attach_raw_tracepoint_with_opts("sys_enter", opts)
         .expect("failed to attach prog");
 
     let map = get_map_mut(&mut obj, "ringbuf");
@@ -1782,8 +1831,6 @@ extern "C" fn uprobe_target() -> usize {
 #[tag(root)]
 #[test]
 fn test_object_uprobe_with_opts() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("uprobe.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__uprobe");
 
@@ -1791,7 +1838,7 @@ fn test_object_uprobe_with_opts() {
     let path = current_exe().expect("failed to find executable name");
     let func_offset = 0;
     let opts = UprobeOpts {
-        func_name: "uprobe_target".to_string(),
+        func_name: Some("uprobe_target".into()),
         ..Default::default()
     };
     let _link = prog
@@ -1807,13 +1854,33 @@ fn test_object_uprobe_with_opts() {
     assert_eq!(result, 1);
 }
 
+#[tag(root)]
+#[test]
+fn test_object_uprobe_with_func_offset() {
+    let mut obj = get_test_object("uprobe.bpf.o");
+    let prog = get_prog_mut(&mut obj, "handle__uprobe");
+
+    let pid = unsafe { libc::getpid() };
+    let path = current_exe().expect("failed to find executable name");
+    let func_offset = get_symbol_offset(&path, "uprobe_target").unwrap();
+    let _link = prog
+        .attach_uprobe_with_opts(pid, path, func_offset, Default::default())
+        .expect("failed to attach prog");
+
+    let map = get_map_mut(&mut obj, "ringbuf");
+    let action = || {
+        let _ = uprobe_target();
+    };
+    let result = with_ringbuffer(&map, action);
+
+    assert_eq!(result, 1);
+}
+
 /// Check that we can attach a BPF program to a uprobe and access the cookie
 /// provided during attach.
 #[tag(root)]
 #[test]
 fn test_object_uprobe_with_cookie() {
-    bump_rlimit_mlock();
-
     let cookie_val = 5u16;
     let mut obj = get_test_object("uprobe.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__uprobe_with_cookie");
@@ -1822,7 +1889,7 @@ fn test_object_uprobe_with_cookie() {
     let path = current_exe().expect("failed to find executable name");
     let func_offset = 0;
     let opts = UprobeOpts {
-        func_name: "uprobe_target".to_string(),
+        func_name: Some("uprobe_target".into()),
         cookie: cookie_val.into(),
         ..Default::default()
     };
@@ -1891,8 +1958,6 @@ fn test_object_perf_buffer_raw() {
     use memmem::Searcher;
     use memmem::TwoWaySearcher;
 
-    bump_rlimit_mlock();
-
     let cookie_val = 42u16;
     let mut obj = get_test_object("tracepoint.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__tracepoint_with_cookie_pb");
@@ -1902,7 +1967,7 @@ fn test_object_perf_buffer_raw() {
         ..TracepointOpts::default()
     };
     let _link = prog
-        .attach_tracepoint_with_opts("syscalls", "sys_enter_getpid", opts)
+        .attach_tracepoint_with_opts(TracepointCategory::Syscalls, "sys_enter_getpid", opts)
         .expect("failed to attach prog");
 
     let map = get_map_mut(&mut obj, "pb");
@@ -1928,8 +1993,6 @@ fn test_object_perf_buffer_raw() {
 #[tag(root)]
 #[test]
 fn test_map_pinned_status() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("map_auto_pin.bpf.o");
     let map = get_map_mut(&mut obj, "auto_pin_map");
     let is_pinned = map.is_pinned();
@@ -1938,15 +2001,13 @@ fn test_map_pinned_status() {
     let get_path = map.get_pin_path().expect("get map pin path failed");
     assert_eq!(expected_path, get_path.to_str().unwrap());
     // cleanup
-    let _ = fs::remove_file(expected_path);
+    let _unused = fs::remove_file(expected_path);
 }
 
-/// Change the root_pin_path and see if it works.
+/// Change the `root_pin_path` and see if it works.
 #[tag(root)]
 #[test]
 fn test_map_pinned_status_with_pin_root_path() {
-    bump_rlimit_mlock();
-
     let obj_path = get_test_object_path("map_auto_pin.bpf.o");
     let mut obj = ObjectBuilder::default()
         .debug(true)
@@ -1964,16 +2025,14 @@ fn test_map_pinned_status_with_pin_root_path() {
     let get_path = map.get_pin_path().expect("get map pin path failed");
     assert_eq!(expected_path, get_path.to_str().unwrap());
     // cleanup
-    let _ = fs::remove_file(expected_path);
-    let _ = fs::remove_dir("/sys/fs/bpf/test_namespace");
+    let _unused = fs::remove_file(expected_path);
+    let _unused = fs::remove_dir("/sys/fs/bpf/test_namespace");
 }
 
 /// Check that we can get program fd by id and vice versa.
 #[tag(root)]
 #[test]
 fn test_program_get_fd_and_id() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("runqslower.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__sched_wakeup");
     let prog_fd = prog.as_fd();
@@ -1985,8 +2044,6 @@ fn test_program_get_fd_and_id() {
 #[tag(root)]
 #[test]
 fn test_map_autocreate_disable() {
-    bump_rlimit_mlock();
-
     let mut open_obj = open_test_object("map_auto_pin.bpf.o");
     let mut auto_pin_map = open_obj
         .maps_mut()
@@ -1999,12 +2056,10 @@ fn test_map_autocreate_disable() {
     open_obj.load().expect("failed to load object");
 }
 
-/// Check that we can resize a map.
+/// Check that we can adjust a map's value size.
 #[tag(root)]
 #[test]
-fn test_map_resize() {
-    bump_rlimit_mlock();
-
+fn test_map_adjust_value_size() {
     let mut open_obj = open_test_object("map_auto_pin.bpf.o");
     let mut resizable = open_obj
         .maps_mut()
@@ -2021,12 +2076,43 @@ fn test_map_resize() {
     assert_eq!(new_len, len * 2);
 }
 
+/// Check that we can adjust a map's maximum entries.
+#[tag(root)]
+#[test]
+fn test_object_map_max_entries() {
+    let mut obj = open_test_object("runqslower.bpf.o");
+
+    // resize the map to have twice the number of entries
+    let mut start = obj
+        .maps_mut()
+        .find(|map| map.name() == OsStr::new("start"))
+        .expect("failed to find `start` map");
+    let initial_max_entries = start.max_entries();
+    let new_max_entries = initial_max_entries * 2;
+    start
+        .set_max_entries(new_max_entries)
+        .expect("failed to set max entries");
+    // check that it reflects on the open map
+    assert_eq!(start.max_entries(), new_max_entries);
+
+    // check that it reflects after loading the map
+    let obj = obj.load().expect("failed to load object");
+    let start = obj
+        .maps()
+        .find(|map| map.name() == OsStr::new("start"))
+        .expect("failed to find `start` map");
+    assert_eq!(start.max_entries(), new_max_entries);
+
+    // check that it reflects after recreating the map handle from map id
+    let start = MapHandle::from_map_id(start.info().expect("failed to get map info").info.id)
+        .expect("failed to get map handle from id");
+    assert!(start.max_entries() == new_max_entries);
+}
+
 /// Check that we are able to attach using ksyscall
 #[tag(root)]
 #[test]
 fn test_attach_ksyscall() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("ksyscall.bpf.o");
     let prog = get_prog_mut(&mut obj, "handle__ksyscall");
     let _link = prog
@@ -2050,8 +2136,6 @@ fn test_attach_ksyscall() {
 #[tag(root)]
 #[test]
 fn test_run_prog_success() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("run_prog.bpf.o");
     let prog = get_prog_mut(&mut obj, "test_1");
 
@@ -2077,11 +2161,57 @@ fn test_run_prog_success() {
 #[tag(root)]
 #[test]
 fn test_run_prog_fail() {
-    bump_rlimit_mlock();
-
     let mut obj = get_test_object("run_prog.bpf.o");
     let prog = get_prog_mut(&mut obj, "test_2");
 
     let input = ProgramInput::default();
     let _err = prog.test_run(input).unwrap_err();
 }
+
+/// Check that we can run a program with `test_run` with `repeat` set.
+///
+/// We set a counter in the program which we bump each time we run the
+/// program.
+/// We check that the counter is equal to the value of `repeat`.
+/// We also check that the duration is non-zero.
+#[tag(root)]
+#[test]
+fn test_run_prog_repeat_and_duration() {
+    let repeat = 100;
+    let payload: [u8; 16] = [
+        0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, // src mac
+        0x11, 0x22, 0x33, 0x44, 0x55, 0x66, // dst mac
+        0x08, 0x00, // ethertype
+        0x00, 0x00, // payload
+    ];
+    let mut obj = get_test_object("run_prog.bpf.o");
+    let prog = get_prog_mut(&mut obj, "xdp_counter");
+
+    let input: ProgramInput<'_> = ProgramInput {
+        data_in: Some(&payload),
+        repeat,
+        ..Default::default()
+    };
+
+    let output = prog.test_run(input).unwrap();
+
+    let map = get_map(&obj, "test_counter_map");
+
+    let counter = map
+        .lookup(&0u32.to_ne_bytes(), MapFlags::ANY)
+        .expect("failed to lookup counter")
+        .expect("failed to retrieve value");
+
+    assert_eq!(output.return_value, libbpf_sys::XDP_PASS);
+    assert_eq!(
+        counter,
+        repeat.to_ne_bytes(),
+        "counter {} != repeat {repeat}",
+        u32::from_ne_bytes(counter.clone().try_into().unwrap())
+    );
+    assert_ne!(
+        output.duration,
+        Duration::ZERO,
+        "duration should be non-zero"
+    );
+}
